
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _TEXTURE_H_
#define _TEXTURE_H_

#include <math.h>

#include <string>
#include <vector>

#include "../shared.src/color.h"
#include "c3ga.h"

class texture {
public:
	texture();
	texture(const std::string &filename);
	texture(const texture &t); ///< copy constructor
	~texture();

	texture &operator=(const texture &t); ///< assignment operator

	void load(const std::string &filename);

	const std::string &get_filename() const;

	float get_grayscale(const c3ga &pt) const {
		return get_color(pt).grayscale();
	}

	float get_grayscale(float u, float v) const {
		return get_color(u, v).grayscale();
	}

	/** Returns the color of the texture at normalized flat point \param pt
	Todo: interpolation!!!!
	*/
	inline color get_color(const c3ga &pt) const {
		return get_color((float)pt[GRADE2][C3GA_E1_NI], (float)pt[GRADE2][C3GA_E2_NI]);
	}

	inline color get_color(float u, float v) const {
		// mod [u, v] coordinates to range [0.0 1.0)
		u = (float)fmod((double)u, 1.0); if (u < 0.0) u += 1.0f;
		v = (float)fmod((double)v, 1.0); if (v < 0.0) v += 1.0f;

		// compute integer & subpixel texture coordinates
		int iu = (int)(u * (float)m_width); // integer u
		float ru = u * (float)m_width - (float)iu; // residue u
		int iv = (int)(v * (float)m_height); // integer v
		float rv = v * (float)m_height - (float)iv; // residue v

		// compute interpolation factors 'a', 'b', 'c' and 'd'
		float a = (1.0f - ru) * (1.0f - rv) / 255.0f; // 0, 0
		float b = ru * (1.0f - rv) / 255.0f; // 1, 0
		float c = (1.0f - ru) * rv / 255.0f; // 0, 1
		float d = ru * rv / 255.0f; // 1, 1

		// compute interpolated texture value
		int idx, i;
		if ((iu == m_width - 1) || (iv == m_height - 1)) {
			float rgba[4];
			// slow special boundary cases
			for (i = 0; i < 3; i++) {
				rgba[i] = (float)m_pixel[((iv) * m_width + (iu)) * 3 + i] * a +
							(float)m_pixel[((iv) * m_width + ((iu+1) % m_width)) * 3 + i] * b +
							(float)m_pixel[(((iv+1) % m_height) * m_width + (iu)) * 3 + i] * c +
							(float)m_pixel[(((iv+1) % m_height) * m_width + ((iu+1) % m_width)) * 3 + i] * d;
			}
			rgba[3] = 1.0;
			return color(rgba);
		}
		else {
			// fast ordinary case
			idx = (iv * m_width + iu) * 3;
			return color((float)m_pixel[idx + 0] * a +
						(float)m_pixel[idx + 3 + 0] * b +
						(float)m_pixel[idx + m_width * 3 + 0] * c +
						(float)m_pixel[idx + m_width * 3 + 3 + 0] * d,

						(float)m_pixel[idx + 1] * a +
						(float)m_pixel[idx + 3 + 1] * b +
						(float)m_pixel[idx + m_width * 3 + 1] * c +
						(float)m_pixel[idx + m_width * 3 + 3 + 1] * d,

						(float)m_pixel[idx + 2] * a +
						(float)m_pixel[idx + 3 + 2] * b +
						(float)m_pixel[idx + m_width * 3 + 2] * c +
						(float)m_pixel[idx + m_width * 3 + 3 + 2] * d,
						
						1.0f);
		}
	}

	float get_bump_map_value(float  u,  float v) const {
			return get_grayscale(u, v) / 1000.0f; // 1000.0f is just an arbitrary constant
	}

	 inline void get_bump_map_value(const c3ga &pt, float bump[3]) const {
		float u = (float)pt[GRADE2][C3GA_E1_NI];
		float v = (float)pt[GRADE2][C3GA_E2_NI];
		const float step = 0.001f;
		float mv[3] = {
			 get_bump_map_value(u, v),
			 get_bump_map_value(u+step, v),
			 get_bump_map_value(u, v+step)
		};

		bump[0] = ((mv[1] - mv[0]) / step);
		bump[1] = ((mv[2] - mv[0]) / step);
		bump[2] = 1.0;
	}


	/// Returns the average color of this texture
	const color &get_average_color() const;



	/// Returns m_valid
	inline bool valid() const {
		return m_valid;
	}

	/// Returns valid() && m_enabled
	inline bool enabled() const {
		return valid() && m_enabled;
	}

	/// sets m_enabled
	void set_enabled(bool e);

	/// Turns this object into a string (e.g., for storing a scene to file)
	//std::string to_string() const;

private:
	/**
	Set to true if texture had been succesfully loaded from file.
	As long as this texture is not valid, m_width, m_height,
	m_average_color and m_pixel are not valid.
	*/
	bool m_valid; 

	/// set to true if this texture is enabled
	bool m_enabled;

	/// The file the texture was read from
	std::string m_filename;

	/// width & height of texture
	int m_width, m_height;

	/// rgb average color of this texture (used for OpenGL visualization)
	color m_average_color;

	std::vector<unsigned char> m_pixel; /// rgb pixels

	/// Scalar: for range [0.0 .... m_magnitude];
//	GAIM_FLOAT m_magnitude;
};

#endif /* _TEXTURE_H_ */
