
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/*
This file contains rather dull and ancient code to parse .rtm (ray tracer model) files.
*/

#include "scene.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "c3ga_util.h"

#include "../shared.src/fltk_stuff.h"
#include "../shared.src/string_util.h"

using namespace c3ga;

void scene::load_rtm(const std::string &filename) {
	printf("Load %s\n", filename.c_str());

	const int maxLineLength = 512;
	char line[maxLineLength];
	int lineNumber = 0, nTabs, l;
	int err = 0;
	FILE *F;
	char entryname[256];
	int i1, vtxIdx[3];
	float f1, f2;
	float red, green, blue;

	const int path_len = 1024;
	char s1[path_len], s2[path_len];
	char path[path_len];
	path_from_filename(path, path_len, filename.c_str());

	model Model;
	mesh Mesh;

//	printf("Load %s\n", filename.c_str());
	
	if ( (F = fopen(filename.c_str(), "rb")) == NULL) {
		throw "scene::load_rtm(): can not open '" + filename + "'";
	}

	while (!feof(F)) {
		if ( (l = read_line(F, line, maxLineLength, &nTabs)) < 0) {
			break;
		}
		lineNumber++;
		if (empty_line(line)) continue;

		if (sscanf(line, "%s ", entryname) != 1) continue;

		if (!strcmpnc(entryname, "number_of_faces")) {		// number_of_faces
			if ((sscanf(line, "%s %d", entryname, &i1) != 2) || (i1 < 0)) {err = -1; break;}
			Mesh.set_nb_faces(i1);
		}

		else if (!strcmpnc(entryname, "number_of_vertices")) {		// number_of_vertices
			if ((sscanf(line, "%s %d", entryname, &i1) != 2) || (i1 < 0)) {err = -1; break;}
			Mesh.set_nb_vertices(i1);
		}

		else if (!strcmpnc(entryname, "face")) {		// face
			if (sscanf(line, "%s %d vtx %d %d %d", entryname, &i1, vtxIdx + 0, vtxIdx + 1, vtxIdx + 2) != 5) {err = -1; break;}
			Mesh.set_face(i1, face(vtxIdx));
		}

		else if (!strcmpnc(entryname, "vertex")) {		// vertex
			float x, y, z, nx, ny, nz, u, v;
			if ((sscanf(line, "%s %d pos %f %f %f normal %f %f %f uv %f %f", 
				entryname, &i1, &x, &y, &z, 
				&nx, &ny, &nz, &u, &v) != 10)) {err = -1;break;}

//			printf("pt = %s,\n", c3ga::string(cga_point(_vectorE3GA(x * e1 + y * e2 + z * e3)), "%2.20e"));
//			printf("att = %s,\n", c3ga::string(_freeBivector(undual((nx * e1 + ny * e2 + nz * e3) ^ ni)), "%2.20e"));
			Mesh.set_vertex(i1, 
				vertex(
    // x, y, z are floats giving the position of the vertex:
					cga_point(_vectorE3GA(x * e1 + y * e2 + z * e3)),
    // nx, ny and nz are floats.
    // They represent the classical surface normal.
					_freeBivector(undual((nx * e1 + ny * e2 + nz * e3) ^ ni)),
	// texture point: 
					_normalizedFlatPoint2D(((u * e1 + v * e2) ^ ni) + noni) ) );
		}

		else if (!strcmpnc(entryname, "ambient_color")) {		// ambient_color
			if ((sscanf(line, "%s %f %f %f", entryname, &red, &green, &blue) != 4)) {err = -1;break;	}
			Model.set_ambient_color(color(red, green, blue));
		}

		else if (!strcmpnc(entryname, "diffuse_color")) {		// diffuse_color
			if ((sscanf(line, "%s %f %f %f", entryname, &red, &green, &blue) != 4)) {err = -1;break;	}
			Model.set_diffuse_color(color(red, green, blue));
		}

		else if (!strcmpnc(entryname, "specular_color")) {		// specular_color
			if ((sscanf(line, "%s %f %f %f", entryname, &red, &green, &blue) != 4)) {err = -1;break;	}
			Model.set_specular_color(color(red, green, blue));
		}

		else if (!strcmpnc(entryname, "shininess")) {		// shininess
			if ((sscanf(line, "%s %f", entryname, &f1) != 2)) {err = -1;break;	}
			Model.set_shininess(f1); 
		}
		else if (!strcmpnc(entryname, "bumpyness")) {		// bumpyness
			if ((sscanf(line, "%s %f", entryname, &f1) != 2)) {err = -1;break;	}
			Model.set_bumpyness(f1); 
		}

		else if ((!strcmpnc(entryname, "ambient_texture")) ||
			(!strcmpnc(entryname, "diffuse_texture")) ||
			(!strcmpnc(entryname, "specular_texture")) ||
			(!strcmpnc(entryname, "shininess_texture")) ||
			(!strcmpnc(entryname, "bumpmap"))) {		// XXX_texture
			if (err = read_filename(line, s1)) break;
			if (strlen(s1)) {
				fl_filename_absolute_to(s2, path_len, s1, path);
				texture &tex(Model.get_texture(entryname));
				tex.load(s2);
				if (tex.valid()) tex.set_enabled(true);
			}
		}

		else if (!strcmpnc(entryname, "direct_lighting")) {		// direct_lighting
			if ((sscanf(line, "%s %f", entryname, &f1) != 2)) {err = -1;break;	}
			Model.set_direct_lighting_factor(f1);
		}
		else if (!strcmpnc(entryname, "reflection")) {		// reflection
			if ((sscanf(line, "%s %f", entryname, &f1) != 2)) {err = -1;break;	}
			Model.set_reflection_factor(f1);
		}
		else if (!strcmpnc(entryname, "refraction")) {		// refraction
			if ((sscanf(line, "%s %f index %f", entryname, &f1, &f2) != 3)) {err = -1;break;	}
			Model.set_refraction_factor(f1);
			Model.set_refractive_index(f2);
		}
		else {
			//printf("warning: scene::load_rtm(): Unknown entry '%s' at line '%d'\n", entryname, lineNumber);
		}
	}

	if (err) {
		// error detected:
		char buf[1024];
		sprintf(buf, "scene::load_rtm(): Invalid entry '%s' at line %d\n", entryname, lineNumber);
		throw std::string(buf);
	}

	fclose(F);


	// Initialize the mesh
	Mesh.init();

	// store mesh in cache:
	Mesh.set_filename(filename);
	m_rtm_mesh_cache[filename] = Mesh;

	// store model in cache:
	Model.set_mesh(&(m_rtm_mesh_cache[filename]));
	m_rtm_model_cache[filename] = Model;

}
