
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _UI_H_
#define _UI_H_

// disable some annoying warnings in VC++ due to FLTK (1.1.6)
#ifdef WIN32
#pragma warning( disable : 4311 4312 4244 )
#endif

#include <stdlib.h>
#include <string>
#include <FL/Fl_Window.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Image.H>

#include "classes.h"
#include "scene.h"
#include "main_window.h"
#include "modeling_window.h"
#include "properties_window.h"
#include "render_window.h"

class ui_state {
public:
	ui_state(const scene &S, int argc, char *argv[], std::string resource_dir); ///< this constructor creates the entire user interface
//	ui_state(const ui_state &ui); ///< copy constructor (just throws exception)
	~ui_state(); ///< destructor

	ui_state &operator=(const ui_state &ui); ///< assignment operator (just throws exception)


	void run(); ///< this runs the user interface

	/// This function is called by the global FLTK callback for the menus
	void menu_callback(Fl_Widget *W, void *Data);

	/// This function is called by the global FLTK callback for the buttons
	void button_callback(Fl_Widget *W, long data);

	/// Should be called every time m_scene changes
	inline void scene_changed() {
		m_model_window->redraw();
	}

	/// Returns the scene current being modeled (const)
	inline const scene &get_scene() const {return m_scene;}
	/// Returns the scene current being modeled
	inline scene &get_scene() {return m_scene;}

	/// Sets the key that is currently being pushed (used for quick 'mode change')
	void set_key_down(int k);
	/// Resets the key that is currently being pushed to none (0) (used for quick 'mode change')
	void set_no_key_down();
	/// Returns the key currently being pushed
	int get_key_down() const;

	/// Returns the mouse mode (depends on m_mouse_mode and m_key_down!)
	mouse_modes get_mouse_mode();

	/*
	Called by main window when its deconstructor is called.
	Prepares for closing of entire user interface
	*/
	void main_window_closed();

	/// Called by scene when some model is selected (updates properties window)
	void update_selected_model();

	/// Called by scene when some light source is selected (updates properties window)
	void update_selected_light_source();

	/// Called by scene when no object is selected (updates properties window)
	void update_selected_none();

	/// Called when the visibility of the properties window is changed (updates the menu check button)
	void set_properties_window_shown(bool s);

	/// Called when the current camera of the scene changes (updates the menu radio buttons)
	void update_current_camera();

	/// Called when 'display lights' in preview mode changes
	void update_display_lights();

	/// must be called when an entirely new scene is loaded into m_scene (updates UI elements and such)
	void new_scene_loaded();

	/// for the renderer, disables modeling interface until render is done
	void disable_modeling();
	/// for the renderer, enables modeling interface when render is done
	void enable_modeling();

private:
	int m_key_down; ///< key that is currently being pressed

	Fl_Widget *m_mouse_mode_button; ///< FLTK button connected to current mouse mode
	mouse_modes m_mouse_mode; ///< the current mouse mode 

	/**
	Returns the button (from m_buttons) associated with \param k
	*/
	Fl_Button *get_button_for_key(int k); 

	/// The main window (containing the entire UI, except for properties)
	Fl_Window *m_main_window;

	/// The OpenGL window that displays the scene
	modeling_window *m_model_window;

	/// The window for controlling properties of models / lights (created once, shown / hidden after that)
	properties_window *m_properties_window;

	/// The window for rendering (created once, shown / hidden after that)
	render_window *m_render_window;

	Fl_Group *m_buttons_group;

	static const int m_nb_buttons = 11;
	Fl_Button *m_buttons[m_nb_buttons];

	Fl_Menu_Bar *m_menu_bar;

	/**
	Images for the buttons:
	*/
	Fl_RGB_Image *m_scale_image;
	Fl_RGB_Image *m_rotate_image;
	Fl_RGB_Image *m_translate_parallel_image;
	Fl_RGB_Image *m_translate_orthogonal_image;

	Fl_RGB_Image *m_camera_fov_image;
	Fl_RGB_Image *m_camera_rotate_image;
	Fl_RGB_Image *m_camera_orbit_image;
	Fl_RGB_Image *m_camera_translate_parallel_image;
	Fl_RGB_Image *m_camera_translate_orthogonal_image;

	Fl_RGB_Image *m_spotlight_image;

	Fl_RGB_Image *m_select_image;

	scene m_scene; ///< the scene currently being modeled

	void create_menu(); ///< creates the menu bar
	void create_button_group(); ///< creates the buttons (translate, select, etc)

	void add_model(); ///< open UI dialog for .rtm file

	/// used intenally by disable_modeling
	bool m_properties_window_was_visible;
};

/// global pointer to ui state
extern ui_state *g_ui_state;

#endif /* _UI_H_ */

