
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <string>

// disable some annoying warnings in VC++ due to FLTK (1.1.6)
#ifdef WIN32
#pragma warning( disable : 4311 4312 4244 )
#endif

#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Menu_Button.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/fl_ask.H>
#include <FL/Fl_File_Chooser.H>

#include "ui.h"
#include "../shared.src/fltk_stuff.h"

namespace { // anonymous namespace for menu code 



enum ui_menu_const {
	// file:
	MI_LOAD_SCENE = 1,
	MI_SAVE_SCENE,
	MI_EXIT,

	// edit:
	MI_ADD_MODEL,
	// todo: add light
	MI_DUPLICATE_MODEL, // todo: duplicate object?
	MI_REMOVE_OBJECT,

	// view:
	MI_VIEW_MODEL_CAMERA,
	MI_VIEW_RENDER_CAMERA,

	MI_VIEW_PROPERTIES,
	MI_VIEW_LIGHTS,

	// render
	MI_RENDER,

	// about
	MI_ABOUT
};

void ui_menu_callback(Fl_Widget *W, void *Data) {
	g_ui_state->menu_callback(W, Data);
}


/**
Menu definition
*/
Fl_Menu_Item ui_menu_table[] = {
  {"File",0,0,0,FL_SUBMENU},
    {"Load scene", 0, ui_menu_callback, (void*)MI_LOAD_SCENE, 0},
    {"Save scene", 0, ui_menu_callback, (void*)MI_SAVE_SCENE, 0},
    {"Exit", 0, ui_menu_callback, (void*)MI_EXIT, 0},
    {0},
  {"Edit",0,0,0,FL_SUBMENU},
    {"Add model", 0, ui_menu_callback, (void*)MI_ADD_MODEL, 0},
	{"Duplicate selected model",	0, ui_menu_callback, (void*)MI_DUPLICATE_MODEL, 0},
    {"Remove selected object", 0, ui_menu_callback, (void*)MI_REMOVE_OBJECT, 0},
    {0},
  {"View",0,0,0,FL_SUBMENU},
	{"Render camera viewpoint",	0, ui_menu_callback, (void*)MI_VIEW_MODEL_CAMERA, FL_MENU_RADIO | FL_MENU_VALUE},
	{"Model camera viewpoint",	0, ui_menu_callback, (void*)MI_VIEW_RENDER_CAMERA, FL_MENU_RADIO | FL_MENU_DIVIDER},

	{"Properties",	0, ui_menu_callback, (void*)MI_VIEW_PROPERTIES, FL_MENU_TOGGLE},
	{"Lights",	0, ui_menu_callback, (void*)MI_VIEW_LIGHTS, FL_MENU_TOGGLE | FL_MENU_VALUE},

    {"About", 0, ui_menu_callback, (void*)MI_ABOUT, 0},
    {0},
  {"Render",0,0,0,FL_SUBMENU},
	{"Render current scene",	0, ui_menu_callback, (void*)MI_RENDER, 0},
    {0},
  {0}
};

/// returns the index of menu item (e.g. MI_SAVE_RENDER)
int get_menu_item(int item) {
	int mi = 0, d = 1;
	while (d) {
		if (ui_menu_table[mi].text == 0) d--;
		else if (ui_menu_table[mi].flags & FL_SUBMENU) d++;
		if (ui_menu_table[mi].user_data_ == (void*)item) return mi;
		mi++;
	}
	// should never happen
	throw std::string("get_menu_item(): unknown menu item");
}

}
void ui_state::create_menu() {
	if (m_main_window == NULL)
		throw std::string("ui_state::create_menu(): main window must be valid");

	m_menu_bar = new Fl_Menu_Bar (0,0, m_main_window->w(), 30); 
	m_menu_bar->menu(ui_menu_table);
	m_menu_bar->global();

}

void ui_state::menu_callback(Fl_Widget *W, void *Data) {
	// find out what the index of the menu item is:
	int item = (int)((unsigned char*)Data - (unsigned char*)NULL);
	
	int mi = get_menu_item(item);

	if (item == MI_ADD_MODEL) {
		add_model();
		scene_changed();
	}
	else if (item == MI_REMOVE_OBJECT) {
		get_scene().remove_selected_object();
		scene_changed();
	}
	else if (item == MI_DUPLICATE_MODEL) {
		try {
			get_scene().duplicate_model(get_scene().get_selected_model());
			scene_changed();
		} catch (std::string &) {/* if model was not selected */}
	}
	else if (item == MI_EXIT) {
		main_window_closed();
		return;
	}
	else if (item == MI_SAVE_SCENE) {
		char *filename;
		if (filename = fl_file_chooser("Save scene", "*.rsn", NULL)) {
			FILE *F = fopen(filename, "w+");
			if (F == NULL)
				fl_alert("Could not open '%s' for writing", filename);
			else {
				const int l = 1024;
				char path[l];
				path_from_filename(path, l, filename);

				fprintf(F, get_scene().to_string(path).c_str());
				fclose(F);
			}
		}
	}
	else if (item == MI_LOAD_SCENE) {
		char *filename;
		if (filename = fl_file_chooser("Load scene", "*.rsn", NULL)) {
			try {
				scene S(filename);
				m_scene = S;
			} catch (const std::string &S) {
				fl_alert( (std::string("Error while loading ") + filename + "\n" + S).c_str());
			}
			scene_changed();
			new_scene_loaded();
		}
	}
	else if (item == MI_VIEW_PROPERTIES) {
		set_properties_window_shown(true);
		m_properties_window->show();
	}
	else if (item == MI_VIEW_MODEL_CAMERA) {
		try {
			get_scene().set_current_camera(0);
			scene_changed();
		} catch (std::string &) {/* if model was not selected */}
	}
	else if (item == MI_VIEW_RENDER_CAMERA) {
		try {
			get_scene().set_current_camera(1);
			scene_changed();
		} catch (std::string &) {/* if model was not selected */}
	}

	else if (item == MI_VIEW_LIGHTS) {
		get_scene().set_display_lights(!get_scene().display_lights());
		update_display_lights();
	}

	else if (item == MI_RENDER) {
		m_render_window->show();
	}

}

void ui_state::set_properties_window_shown(bool s) {
	int mi = get_menu_item(MI_VIEW_PROPERTIES);

	ui_menu_table[mi].flags = FL_MENU_TOGGLE | ((s) ? FL_MENU_VALUE : 0);	
}

void ui_state::update_current_camera() {
	int idx = get_scene().get_current_camera_idx();
	int mi_m = get_menu_item(MI_VIEW_MODEL_CAMERA);
	int mi_r = get_menu_item(MI_VIEW_RENDER_CAMERA);

	ui_menu_table[mi_m].flags = FL_MENU_RADIO | ((idx == 0) ? FL_MENU_VALUE : 0);	
	ui_menu_table[mi_r].flags = FL_MENU_RADIO | ((idx == 1) ? FL_MENU_VALUE : 0) | FL_MENU_DIVIDER;	
}

void ui_state::update_display_lights() {
	int mi = get_menu_item(MI_VIEW_LIGHTS);
	ui_menu_table[mi].flags = FL_MENU_TOGGLE | ((get_scene().display_lights()) ? FL_MENU_VALUE : 0);

}
