
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _C3GA_UTIL_H_
#define _C3GA_UTIL_H_

#include <math.h>
#include <string>
#include "c3ga.h"


namespace c3ga {
/// Returns a point at the location specified by 3D Euclidean vector \param p
inline normalizedPoint cga_point(const vectorE3GA &v) {
	return _normalizedPoint(v + no + 0.5 * norm_e2(v) * ni);
}

/// Returns a point at the location specified by 2D Euclidean vector \param p
inline normalizedPoint cga_point(const vectorE2GA &v) {
	return _normalizedPoint(v + no + 0.5 * norm_e2(v) * ni);
}

/// Returns a point at the location specified by 2D Euclidean vector \param p
inline normalizedPoint cga_point(const flatPoint &fpt) { 
	return cga_point(_vectorE3GA(no << (fpt * (1.0 / _Float(fpt & noni)))));
}

/**
Returns the Euclidean distance between two points, squared.
Assumes the points are 'normalized'
*/
inline mv::Float flat_point_distance_squared(const flatPoint &fpt1, const flatPoint &fpt2) {
	return _Float(norm_e2(fpt1 - fpt2));
}

inline mv::Float flat_point_distance_squared(const normalizedFlatPoint &fpt1, const normalizedFlatPoint &fpt2) {
	return _Float(norm_e2(fpt1 - fpt2));
}

/**
Returns the Euclidean distance between two points.
Assumes the points are 'normalized'
*/
inline mv::Float flat_point_distance(const flatPoint &fpt1, const flatPoint &fpt2) {
	return (mv::Float)sqrt(flat_point_distance_squared(fpt1, fpt2));
}

inline mv::Float flat_point_distance(const normalizedFlatPoint &fpt1, const normalizedFlatPoint &fpt2) {
	return (mv::Float)sqrt(flat_point_distance_squared(fpt1, fpt2));
}

mv exp(const mv &x, int order = 9);

/// Multiplies the current OpenGL matrix with rotor \param v
void unit_versor_glMult(const EXForm &v);

/// dissects a point pair \param pp into two CGA points \param p1 and \param p2 .
void dissect_point_pair(const pointPair &pp, point &pt1, point &pt2);

/// dissects a point pair \param pp into two CGA points \param p1 and \param p2 .
void dissect_point_pair_normalized(const pointPair &pp, normalizedPoint &pt1, normalizedPoint &pt2);

/// dissects a point pair \param pp into two normalized flat points \param fp1 and \param fp2 .
void dissect_point_pair_normalized_flat(const pointPair &pp, normalizedFlatPoint &fpt1, normalizedFlatPoint &fpt2);

/** parses a string, returns an mv (or throws -1 on failure) 
	(\param startDL and \param endDL are start and end delimiters,
	can be -1 for none)
	*/
mv parseMVString(const std::string &str, int startDL=-1, int endDL=-1);

/// Extracts location (normalized point) from tangent vector
inline normalizedPoint get_pt(const tangentVector &tv) {
	return _normalizedPoint(-tv * inverse(ni << tv));
} 

/// Extracts location (normalized point) from tangent bivector
inline normalizedPoint get_pt(const tangentBivector &tbv) {
	return _normalizedPoint(-tbv * inverse(ni << tbv));
} 

/// Extracts attitude (free bivector)  from tangent vector
inline freeVector get_att(const tangentVector &tv) {
	return _freeVector(-(ni  << tv) ^ ni);
}

/// Extracts attitude (free bivector)  from tangent bivector
inline freeBivector get_att(const tangentBivector &tbv) {
	return _freeBivector(-(ni  << tbv) ^ ni);
}


} // end of namespace c3ga

#endif /* #ifndef _C3GA_UTIL_H_ */
