
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _FACE_H_
#define _FACE_H_

#include "classes.h"

class face {
public:
	face(); ///< null constructor
	/** constructor
	Sets m_vtx_idx to \param vtx_idx
	precompute_constants must still be called with the appropriate mesh argument
	to set all the members of this class
	*/
	face(const int vtx_idx[3]); 

	face(const face &f); ///< copy constructor
	~face(); ///< destructor


	face &operator=(const face &f); //< copy operator

	/**
	Called when a model is first loaded. Computes the following constants:
	m_edgeLine, m_pl, m_u2, m_v2, m_w2, m_uv, m_vw, m_wu, m_u_w2, m_uvDir

	m_vtxIdx must have been set before calling this function.
	*/
	void precompute_constants(const mesh &m);

	/// Retrieves an element of m_vtx_idx; \param idx must be 0<=idx<=2; array bounds not checked
	inline int get_vtx_idx(int idx) const {return m_vtx_idx[idx];}

	/// Retrieves vertex \param idx given the mesh \param m; \param idx must be 0<=idx<=2; array bounds not checked
	const vertex &get_vtx(int idx, const mesh &m) const;

	/// Returns whether this face depends on vertex \param idx
	bool uses_vtx_idx(int vtx_idx) const;

	/// Computes m_uv_dir 
	void compute_local_texture_basis(const mesh &m);

	/**
	\param pt is the location of the ray
	\param dir is the direction of the ray
	\param spt is where the intersection info goes
	\param find_closest instructs the function to compare and overwrite spt only if further away
	*/
	bool find_intersection(const p3ga::normalizedPoint &pt, const p3ga::vectorE3GA &dir, surface_point &spt, bool find_closest) const;

	/**
	Computes the details for surface point \param spt .
	model \param m is required for vertex and bumpmap lookup
	*/
	void compute_surface_point_details(surface_point &spt, const model &m) const;

	int get_valid() const {return m_valid;}
private:
	int m_vtx_idx[3];	///< indices for each vertex. Indices point into the parent mesh vertex array

	/** One plane for each edge
	 Each plane contains the edge, and is orthogonal to the face
	 These are used to determine if a line intersect the triangle or not
	 */
	p3ga::plane m_edgePl[3];
	p3ga::vectorE3GA m_edgeDir[3];

	p3ga::plane m_pl; ///< the plane in homogenous model of the face

	/// precomputed constants for interpolation:
	p3ga::mv::Float m_u2, m_v2, m_w2, m_u_v, m_v_w, m_w_u, m_u_w2;

	/// docs todo
	p3ga::mv::Float m_uv_dir[2][2]; 


	int m_valid; // 0 if the surface area of this plane is 0


	/**
	For internal use.
	Called by compute_surface_point_details()

	Computes the interpolation weights for vertices 0, 1, 2
	for the (normalized) point \param pt .
	\param vtx0 , \param vtx1 and \param vtx2 are the
	vertices of this face.
	Returns them in \param iw .
	*/
	void compute_interpolation_weights(const p3ga::normalizedPoint &pt, p3ga::mv::Float iw[3], const vertex &vtx0) const;



};

#endif /* _FACE_H_ */
