#ifndef _P3LA_H_
#define _P3LA_H_

#include <stdio.h>
#include <math.h>

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

#ifndef GAIM_FLOAT
#define GAIM_FLOAT double 
#endif 


class p3la_vec;
class p3la_point;
class p3la_line;
class p3la_plane;
class p3la_tm; // tm = transformation matrix

// inner product of vector and point
GAIM_FLOAT operator%(const p3la_vec &v, const p3la_point &pt);
inline GAIM_FLOAT operator%(const p3la_point &pt, const p3la_vec &v) {return operator%(v, pt);}

class p3la_vec {
public:
	p3la_vec() : x(0.0), y(0.0), z(0.0) {}; // null constructor
	p3la_vec(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) : x(ax), y(ay), z(az) {}; // parametarised constructor
	p3la_vec(const GAIM_FLOAT c[3]) : x(c[0]), y(c[1]), z(c[2]) {}; // parametarised constructor
	p3la_vec(const p3la_vec &v) : x(v.x), y(v.y), z(v.z) {}; // copy constructor
	p3la_vec(const p3la_point &pt);
	~p3la_vec() {};

	inline void null() {x = y = z = 0.0;}

	p3la_vec& operator=(const p3la_vec &v) { // assignment operator
		x = v.x; y = v.y; z = v.z;
		return *this;
    }

	p3la_vec operator^(const p3la_vec &v) const;
	p3la_vec operator+(const p3la_vec &v) const;
	p3la_vec operator-(const p3la_vec &v) const;
	inline p3la_vec operator-() const {return p3la_vec(-x, -y, -z);}

	inline int set(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) {x = ax; y = ay; z = az; return 0;};
	inline int set(const GAIM_FLOAT c[3]) {x = c[0]; y = c[1]; z = c[2]; return 0;};
	inline int set(const p3la_vec &v)  {x = v.x; y = v.y; z = v.z; return 0;};
	int set(const p3la_point &pt);


	inline p3la_vec normalize() const {
		GAIM_FLOAT s = sqrt(x * x + y * y + z * z);
		return p3la_vec(x / s, y / s, z / s);
	}
	inline GAIM_FLOAT length() const {return sqrt(x * x + y * y + z * z);};
	inline GAIM_FLOAT magnitude() const {return length();};

	void print(const char *text, const char *fmt = NULL) const;

	GAIM_FLOAT x, y, z; // the coordinates
};

inline p3la_vec operator*(const p3la_vec &v, GAIM_FLOAT s) {return p3la_vec(v.x * s, v.y * s, v.z * s);}
inline p3la_vec operator*(GAIM_FLOAT s, const p3la_vec &v) {return operator*(v, s);}
inline p3la_vec operator/(const p3la_vec &v, GAIM_FLOAT s) {return p3la_vec(v.x / s, v.y / s, v.z / s);}

class p3la_point {
public:
	p3la_point() : x(0.0), y(0.0), z(0.0), w(0.0) {}; // null constructor
	p3la_point(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) : x(ax), y(ay), z(az), w(1.0) {}; // parametarised constructor
	p3la_point(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az, GAIM_FLOAT aw) : x(ax), y(ay), z(az), w(aw) {}; // parametarised constructor
	p3la_point(const GAIM_FLOAT c[3]) : x(c[0]), y(c[1]), z(c[2]), w(1.0) {}; // parametarised constructor
	p3la_point(const p3la_point &pt) : x(pt.x), y(pt.y), z(pt.z), w(pt.w) {}; // copy constructor
	p3la_point(const p3la_vec &v) : x(v.x), y(v.y), z(v.z), w(1.0) {};
	~p3la_point() {};

	inline void null() {x = y = z = w = 0.0;}

	p3la_point& operator=(const p3la_point &pt) { // assignment operator
		x = pt.x; y = pt.y; z = pt.z; w = pt.w;
		return *this;
    }

	inline void set(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) {x = ax; y = ay; z = az; w = 1.0;};
	inline void set(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az, GAIM_FLOAT aw) {x = ax; y = ay; z = az; w = aw;};
	inline void set(const GAIM_FLOAT c[3])  {x = c[0]; y = c[1]; z = c[2]; w = 1.0;};
	inline void set(const p3la_point &pt) {operator=(pt);}
	inline void set(const p3la_vec &v) {x = v.x; y = v.y; z = v.z; w = 1.0;}

	p3la_point operator+(const p3la_vec &v) const;
	p3la_point operator-(const p3la_vec &v) const;
	p3la_vec operator-(const p3la_point &pt) const;

	inline p3la_point normalize() const {return p3la_point(x / w, y / w, z / w, w / w);}

	inline GAIM_FLOAT magnitude() const {
		return w;
	}


	void print(const char *text, const char *fmt = NULL) const;

	GAIM_FLOAT x, y, z, w; // the coordinates
};

class p3la_line {
public:
	p3la_line() : xy(0.0), zx(0.0), yz(0.0), xw(0.0), yw(0.0), zw(0.0) {}; // null constructor
	p3la_line(GAIM_FLOAT ayz, GAIM_FLOAT azx , GAIM_FLOAT axy, GAIM_FLOAT axw, GAIM_FLOAT ayw, GAIM_FLOAT azw) : yz(ayz), zx(azx), xy(axy), xw(axw), yw(ayw), zw(azw) {};
	p3la_line(const p3la_line &ln) : yz(ln.yz), zx(ln.zx), xy(ln.xy), xw(ln.xw), yw(ln.yw), zw(ln.zw) {}; // copy constructor
	p3la_line(const p3la_point &pt1, const p3la_point &pt2) {set(pt1, pt2);}; // two point constructor
	p3la_line(const p3la_point &pt, const p3la_vec &v) {set(pt, v);}; // point-direction constructor
	~p3la_line() {};

	inline void null() {yz = zx = xy = xw = yw = zw = 0.0;}

	p3la_line& operator=(const p3la_line &ln) { // assignment operator
		yz = ln.yz; zx = ln.zx; xy = ln.xy;
		xw = ln.xw;	yw = ln.yw;	zw = ln.zw;
		return *this;
    }

	int set(const p3la_point &pt1, const p3la_point &pt2);
	int set(const p3la_point &pt, const p3la_vec &v);

	p3la_point point() const; // returns the point on the line closest to the origin
	p3la_vec direction() const; // returns the direction of the line (unit)
	p3la_vec directionNonUnit() const; // returns the direction of the line (non-unit)

	inline p3la_line normalize() const {
		GAIM_FLOAT s = sqrt(xw * xw + yw * yw + zw * zw);
		return p3la_line(yz / s, zx / s, xy / s, xw / s, yw / s, zw / s);
	}

	inline GAIM_FLOAT magnitude() const {
		return sqrt(xw * xw + yw * yw + zw * zw);
	}


	void print(const char *text, const char *fmt = NULL) const;

	GAIM_FLOAT yz, xy, zx, xw, yw, zw;
};

class p3la_plane {
public:
	p3la_plane() : d(0.0), z(0.0), y(0.0), x(0.0) {}; // null constructor
	p3la_plane(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az, GAIM_FLOAT ad) : x(ax), y(ay), z(az), d(ad) {}; // parametarised constructor (normal + distance) (todo: check if ok...)
	p3la_plane(const GAIM_FLOAT nl[3], GAIM_FLOAT ad) : x(nl[0]), y(nl[1]), z(nl[2]), d(ad) {}; // parametarised constructor (normal + distance) (todo: check if ok...)
	p3la_plane(const p3la_plane &pl) : d(pl.d), z(pl.z), y(pl.y), x(pl.x) {}; // copy constructor
	p3la_plane(const p3la_point &pt1, const p3la_point &pt2, const p3la_point &pt3) {set(pt1, pt2, pt3);}; // three point constructor
	p3la_plane(const p3la_point &pt, const p3la_line &ln) {set(pt, ln);}; // point line constructor
	p3la_plane(const p3la_line &ln, const p3la_point &pt) {set(ln, pt);}; // line point constructor
	p3la_plane(const p3la_vec &v, GAIM_FLOAT ad) : x(v.x), y(v.y), z(v.z), d(ad) {}; // normal constructor
	p3la_plane(const p3la_point &pt, const p3la_vec &v) : x(v.x), y(v.y), z(v.z), d(v % pt) {}; // point & normal constructor
	~p3la_plane() {};

	inline void null() {x = y = z = d = 0.0;}

	p3la_plane& operator=(const p3la_plane &pl) { // assignment operator
		z = pl.z; y = pl.y; x = pl.x, d = pl.d;
		return *this;
    }

	inline p3la_plane& operator+=(const p3la_plane &pl) {d += pl.d; z += pl.z; y += pl.y; x += pl.x; return *this;}

	int set(const p3la_point &pt1, const p3la_point &pt2, const p3la_point &pt3);
	int set(const p3la_point &pt, const p3la_line &ln);
	int set(const p3la_line &ln, const p3la_point &pt);
	inline int set(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az, GAIM_FLOAT ad) {x = ax; y = ay; z = az; d = ad; return 0;}
	inline int set(const p3la_point &pt, const p3la_vec &v) {x = v.x; y = v.y; z = v.z; d  = v % pt; return 0;};
	inline int set(const p3la_vec &v, GAIM_FLOAT ad) {x = v.x; y = v.y; z = v.z; d  = ad; return 0;};

	inline p3la_vec normal() const {return p3la_vec(x, y, z);}

	inline p3la_plane normalize() const {
		GAIM_FLOAT s = sqrt(x * x + y * y + z * z);
		return p3la_plane(x / s, y / s, z / s, d / s);
	}

	inline GAIM_FLOAT magnitude() const {
		return sqrt(x * x + y * y + z * z);
	}


	void print(const char *text, const char *fmt = NULL) const;

	GAIM_FLOAT x, y, z, d; // the coordinates
};

class p3la_tm {
public:
	p3la_tm() : m11(1.0), m12(0.0), m13(0.0), m14(0.0), m21(0.0), m22(1.0), m23(0.0), m24(0.0),
		m31(0.0), m32(0.0), m33(1.0), m34(0.0), m41(0.0), m42(0.0), m43(0.0), m44(1.0) {}; // identity constructor
	p3la_tm(
		GAIM_FLOAT am11, GAIM_FLOAT am12, GAIM_FLOAT am13, GAIM_FLOAT am14,
		GAIM_FLOAT am21, GAIM_FLOAT am22, GAIM_FLOAT am23, GAIM_FLOAT am24,
		GAIM_FLOAT am31, GAIM_FLOAT am32, GAIM_FLOAT am33, GAIM_FLOAT am34,
		GAIM_FLOAT am41, GAIM_FLOAT am42, GAIM_FLOAT am43, GAIM_FLOAT am44) :
		m11(am11), m12(am12), m13(am13), m14(am14),
		m21(am21), m22(am22), m23(am23), m24(am24),
		m31(am31), m32(am32), m33(am33), m34(am34),
		m41(am41), m42(am42), m43(am43), m44(am44) {};
	p3la_tm(const p3la_vec &axis, GAIM_FLOAT angle) {set(axis, angle);}
	p3la_tm(const p3la_vec &tr) {set(tr);}
	~p3la_tm() {};

	inline void identity() {
		m11 = m22 = m33 = m44 = 1.0;
		m12 = m13 = m14 = m21 =
		m23 = m24 = m31 = m32 =
		m34 = m41 = m42 = m43 = 0.0;}

	p3la_tm transpose() const;

	int set(const p3la_vec &axis, GAIM_FLOAT angle);
	int set(const p3la_vec &tr);

	void print(const char *text, const char *fmt = NULL) const;

	GAIM_FLOAT m11, m12, m13, m14;
	GAIM_FLOAT m21, m22, m23, m24;
	GAIM_FLOAT m31, m32, m33, m34;
	GAIM_FLOAT m41, m42, m43, m44;
};

p3la_vec operator*(const p3la_tm &tm, const p3la_vec &pt);
p3la_point operator*(const p3la_tm &tm, const p3la_point &pt);
p3la_line operator*(const p3la_tm &tm, const p3la_line &ln);
p3la_plane operator*(const p3la_tm &tm, const p3la_plane &pl);
p3la_tm operator*(const p3la_tm &tm1, const p3la_tm &tm2);

// signed distance of plane and point
GAIM_FLOAT operator%(const p3la_plane &pl, const p3la_point &pt);
inline GAIM_FLOAT operator%(const p3la_point &pt, const p3la_plane &pl) {return operator%(pl, pt);}

// intersection of line and plane
p3la_point operator%(const p3la_plane &pl, const p3la_line &ln);
inline p3la_point operator%(const p3la_line &ln, const p3la_plane &pl) {return operator%(pl, ln);}

// cosine of angle of two [lines, planes, vectors]
GAIM_FLOAT operator%(const p3la_plane &pl1, const p3la_plane &pl2);
GAIM_FLOAT operator%(const p3la_line &ln1, const p3la_line &ln2);
GAIM_FLOAT operator%(const p3la_vec &v1,const  p3la_vec &v2);

// 'signed distance' of two lines
GAIM_FLOAT side(const p3la_line &ln1, const p3la_line &ln2);


inline p3la_vec::p3la_vec(const p3la_point &pt) : x(pt.x / pt.w), y(pt.y / pt.w), z(pt.z / pt.w) {}; // parametarised constructor

inline int p3la_vec::set(const p3la_point &pt) {
	x = pt.x / pt.w;
	y = pt.y / pt.w;
	z = pt.z / pt.w;
	return 0;
}

inline p3la_vec p3la_vec::operator^(const p3la_vec &v) const {
	return p3la_vec(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x);
}

inline p3la_vec p3la_vec::operator+(const p3la_vec &v) const {
	return p3la_vec(x + v.x, y + v.y, z + v.z);
}

inline p3la_vec p3la_vec::operator-(const p3la_vec &v) const {
	return p3la_vec(x - v.x, y - v.y, z - v.z);
}

inline p3la_point p3la_point::operator+(const p3la_vec &v) const {
	return p3la_point(x + v.x * w, y + v.y * w, z + v.z * w, w);
}

inline p3la_point p3la_point::operator-(const p3la_vec &v) const {
	return p3la_point(x - v.x * w, y - v.y * w, z - v.z * w, w);
}

inline p3la_vec p3la_point::operator-(const p3la_point &pt) const {
	return p3la_vec(x / w - pt.x / pt.w, y / w - pt.y / pt.w, z / w - pt.z / pt.w);
}

inline int p3la_line::set(const p3la_point &pt1, const p3la_point &pt2) {
	yz = - pt1.z * pt2.y + pt1.y * pt2.z ;
	zx = + pt1.z * pt2.x - pt1.x * pt2.z ;
	xy = - pt1.y * pt2.x + pt1.x * pt2.y ;
	xw = - pt1.w * pt2.x + pt1.x * pt2.w ;
	yw = - pt1.w * pt2.y + pt1.y * pt2.w ;
	zw = - pt1.w * pt2.z + pt1.z * pt2.w ;

	return 0;
}

inline int p3la_line::set(const p3la_point &pt, const p3la_vec &v) {
	yz = - pt.z * v.y + pt.y * v.z ;
	zx = + pt.z * v.x - pt.x * v.z ;
	xy = - pt.y * v.x + pt.x * v.y ;
	xw = - pt.w * v.x;
	yw = - pt.w * v.y;
	zw = - pt.w * v.z;
	return 0;
}


inline p3la_point p3la_line::point() const {
	GAIM_FLOAT s = magnitude();
	return p3la_point(p3la_vec(yz / s, zx / s, xy / s) ^ p3la_vec(xw / s, yw / s, zw / s));
}

inline p3la_vec p3la_line::direction() const {
	GAIM_FLOAT s = magnitude();
	return p3la_vec(xw / s, yw / s, zw / s);
}

inline p3la_vec p3la_line::directionNonUnit() const {
	return p3la_vec(xw, yw, zw);
}

inline int p3la_plane::set(const p3la_point &pt1, const p3la_point &pt2, const p3la_point &pt3) {
	p3la_line ln(pt1, pt2);
	set(ln, pt3);

	return 0;
}

inline int p3la_plane::set(const p3la_point &pt, const p3la_line &ln) {
	x = + pt.w * ln.yz - pt.z * ln.yw + pt.y * ln.zw ;
	y = + pt.w * ln.zx + pt.z * ln.xw - pt.x * ln.zw ;
	z = + pt.w * ln.xy - pt.y * ln.xw + pt.x * ln.yw ;
	d = + pt.z * ln.xy + pt.y * ln.zx + pt.x * ln.yz ;

	return 0;
}

inline int p3la_plane::set(const p3la_line &ln, const p3la_point &pt) {
	x = + ln.zw * pt.y - ln.yw * pt.z + ln.yz * pt.w ;
	y = - ln.zw * pt.x + ln.xw * pt.z + ln.zx * pt.w ;
	z = + ln.yw * pt.x - ln.xw * pt.y + ln.xy * pt.w ;
	d = + ln.yz * pt.x + ln.zx * pt.y + ln.xy * pt.z ;
	return 0;
}


inline int p3la_tm::set(const p3la_vec &a, GAIM_FLOAT angle) {
	GAIM_FLOAT sinA, cosA, versA;
	sinA = (GAIM_FLOAT)sin(angle);
	cosA = (GAIM_FLOAT)cos(angle);
	versA = (GAIM_FLOAT)1.0 - cosA;

	m11 = a.x * a.x * versA + cosA;
	m22 = a.y * a.y * versA + cosA;
	m33 = a.z * a.z * versA + cosA;

	m12 = a.x * a.y * versA - a.z * sinA;
	m23 = a.y * a.z * versA - a.x * sinA;
	m31 = a.x * a.z * versA - a.y * sinA;

	m13 = a.x * a.z * versA + a.y * sinA;
	m21 = a.x * a.y * versA + a.z * sinA;
	m32 = a.y * a.z * versA + a.x * sinA;

	m44 = 1.0;
	m14 = m24 = m34 = m41 = m42 = m43 = 0.0;

	return 0;
}

inline int p3la_tm::set(const p3la_vec &tr) {
	m14 = tr.x;
	m24 = tr.y;
	m34 = tr.z;
	m11 = m22 = m33 = m44 = 1.0;

	m12 = m13 =
	m21 = m23 =
	m31 = m32 =
	m41 = m42 = m43 = 0.0;

	return 0;
}

inline p3la_vec operator*(const p3la_tm &tm, const p3la_vec &v) {
	p3la_vec result(
		tm.m11 * v.x + tm.m12 * v.y + tm.m13 * v.z,
		tm.m21 * v.x + tm.m22 * v.y + tm.m23 * v.z,
		tm.m31 * v.x + tm.m32 * v.y + tm.m33 * v.z);
	return result;
}

inline p3la_point operator*(const p3la_tm &tm, const p3la_point &pt) {
	p3la_point result(
		tm.m11 * pt.x + tm.m12 * pt.y + tm.m13 * pt.z + tm.m14 * pt.w,
		tm.m21 * pt.x + tm.m22 * pt.y + tm.m23 * pt.z + tm.m24 * pt.w,
		tm.m31 * pt.x + tm.m32 * pt.y + tm.m33 * pt.z + tm.m34 * pt.w,
		tm.m41 * pt.x + tm.m42 * pt.y + tm.m43 * pt.z + tm.m44 * pt.w);
	return result;
}

inline p3la_line operator*(const p3la_tm &tm, const p3la_line &ln) {
	// seperate the line into two points, transform those, reconstruct line
	GAIM_FLOAT s = ln.magnitude();
	p3la_vec u(ln.xw / s, ln.yw / s, ln.zw / s);
	p3la_point pt1(p3la_vec(ln.yz / s, ln.zx / s, ln.xy / s) ^ u);
	p3la_point pt2(pt1 + u);
	return p3la_line(tm *pt2, tm * pt1);
}

inline p3la_plane operator*(const p3la_tm &tm, const p3la_plane &pl) {
	GAIM_FLOAT x = tm.m11 * pl.x + tm.m12 * pl.y + tm.m13 * pl.z;
	GAIM_FLOAT y = tm.m21 * pl.x + tm.m22 * pl.y + tm.m23 * pl.z;
	GAIM_FLOAT z = tm.m31 * pl.x + tm.m32 * pl.y + tm.m33 * pl.z;
	// assuming that tm contains _only_ rotation and translation, we can do this:
	return p3la_plane(x, y, z,
		pl.d + tm.m14 * x + tm.m24 * y + tm.m34 * z);
}

/*
p3la_plane operator*(const p3la_tm &tm, const p3la_plane &pl) {
	// seperate the plane into a normal vector and a point, transform those, reconstruct plane
	p3la_vec v(pl.x, pl.y, pl.z), w = tm * v;
	p3la_point pt(v * pl.d);
	return p3la_plane(w, (tm * pt) % w);
}*/

inline p3la_tm operator*(const p3la_tm &tm1, const p3la_tm &tm2) {
	p3la_tm result(
		tm1.m11 * tm2.m11 + tm1.m12 * tm2.m21 + tm1.m13 * tm2.m31 + tm1.m14 * tm2.m41,
		tm1.m11 * tm2.m12 + tm1.m12 * tm2.m22 + tm1.m13 * tm2.m32 + tm1.m14 * tm2.m42,
		tm1.m11 * tm2.m13 + tm1.m12 * tm2.m23 + tm1.m13 * tm2.m33 + tm1.m14 * tm2.m43,
		tm1.m11 * tm2.m14 + tm1.m12 * tm2.m24 + tm1.m13 * tm2.m34 + tm1.m14 * tm2.m44,

		tm1.m21 * tm2.m11 + tm1.m22 * tm2.m21 + tm1.m23 * tm2.m31 + tm1.m24 * tm2.m41,
		tm1.m21 * tm2.m12 + tm1.m22 * tm2.m22 + tm1.m23 * tm2.m32 + tm1.m24 * tm2.m42,
		tm1.m21 * tm2.m13 + tm1.m22 * tm2.m23 + tm1.m23 * tm2.m33 + tm1.m24 * tm2.m43,
		tm1.m21 * tm2.m14 + tm1.m22 * tm2.m24 + tm1.m23 * tm2.m34 + tm1.m24 * tm2.m44,

		tm1.m31 * tm2.m11 + tm1.m32 * tm2.m21 + tm1.m33 * tm2.m31 + tm1.m34 * tm2.m41,
		tm1.m31 * tm2.m12 + tm1.m32 * tm2.m22 + tm1.m33 * tm2.m32 + tm1.m34 * tm2.m42,
		tm1.m31 * tm2.m13 + tm1.m32 * tm2.m23 + tm1.m33 * tm2.m33 + tm1.m34 * tm2.m43,
		tm1.m31 * tm2.m14 + tm1.m32 * tm2.m24 + tm1.m33 * tm2.m34 + tm1.m34 * tm2.m44,

		tm1.m41 * tm2.m11 + tm1.m42 * tm2.m21 + tm1.m43 * tm2.m31 + tm1.m44 * tm2.m41,
		tm1.m41 * tm2.m12 + tm1.m42 * tm2.m22 + tm1.m43 * tm2.m32 + tm1.m44 * tm2.m42,
		tm1.m41 * tm2.m13 + tm1.m42 * tm2.m23 + tm1.m43 * tm2.m33 + tm1.m44 * tm2.m43,
		tm1.m41 * tm2.m14 + tm1.m42 * tm2.m24 + tm1.m43 * tm2.m34 + tm1.m44 * tm2.m44
		);
	return result;
}

inline p3la_tm p3la_tm::transpose() const {
	return p3la_tm(m11, m21, m31, m41,
		m12, m22, m32, m42,
		m13, m23, m33, m43,
		m14, m24, m34, m44);
}


inline GAIM_FLOAT operator%(const p3la_plane &pl, const p3la_point &pt) {
	return +pl.x * pt.x + pl.y * pt.y + pl.z * pt.z - pl.d * pt.w;
}

inline GAIM_FLOAT operator%(const p3la_vec &v, const p3la_point &pt) {
	return (v.x * pt.x + v.y * pt.y + v.z * pt.z) / pt.w;
}

inline p3la_point operator%(const p3la_plane &pl, const p3la_line &ln) {
	return p3la_point(
	pl.y * ln.xy -pl.z * ln.zx - pl.d * ln.xw,
	-pl.x * ln.xy +pl.z * ln.yz - pl.d * ln.yw,
	pl.x * ln.zx -pl.y * ln.yz - pl.d * ln.zw,
	-pl.x * ln.xw -pl.y * ln.yw -pl.z * ln.zw);


}

inline GAIM_FLOAT operator%(const p3la_plane &pl1, const p3la_plane &pl2) {
	return pl1.x * pl2.x + pl1.y * pl2.y + pl1.z * pl2.z;
}

inline GAIM_FLOAT operator%(const p3la_line &ln1, const p3la_line &ln2) {
	return ln1.xw * ln2.xw + ln1.yw * ln2.yw + ln1.zw * ln2.zw;
}

inline GAIM_FLOAT operator%(const p3la_vec &v1, const p3la_vec &v2) {
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

inline GAIM_FLOAT side(const p3la_line &ln1, const p3la_line &ln2) {
	return ln1.xy * ln2.zw + ln1.zx * ln2.yw + ln1.yz * ln2.xw +
		ln1.xw * ln2.yz + ln1.yw * ln2.zx + ln1.zw * ln2.xy ;
}


#endif /* _P3LA_H_ */
