
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _INTERSECTION_INFO_H_
#define _INTERSECTION_INFO_H_

#include <stdlib.h>

class model;

/**
This class holds precomputed intersection info.
This info is constructed during a seperate 'rendering' pass.
The actual rendering pass can use this info; then it
does not have to search for intersections, which balances
the geometric operations used during ray tracing (otherwise >90% 
of the rendering time would be spent on searching for intersections).
*/
class intersection_info {
public:
	intersection_info() : m_model(NULL), 
		m_face_idx(-1), m_light_visible(0),
		m_reflection_info(NULL),
		m_refraction_info(NULL) {
	}

	intersection_info(model *m, int face_idx, unsigned int light_visible,
		intersection_info *refl_info, intersection_info *refr_info) :
		m_model(m), 
		m_face_idx(face_idx), m_light_visible(light_visible),
		m_reflection_info(refl_info),
		m_refraction_info(refr_info) {
	}

	intersection_info(intersection_info &I) :  ///< copy constructor
		m_model(I.m_model), 
		m_face_idx(I.m_face_idx), 
		m_light_visible(I.m_light_visible),
		m_reflection_info(new intersection_info(*I.m_reflection_info)),
		m_refraction_info(new intersection_info(*I.m_refraction_info)) {
	}

	// ....
	~intersection_info() {
		// no need to delete model (it is kept by the scene)
		if (m_reflection_info) delete m_reflection_info;
		if (m_refraction_info) delete m_refraction_info;
	}

	/// assignment operator
	intersection_info &operator=(intersection_info &I) {
		if (this != &I) {
			m_model = I.m_model;
			m_face_idx = I.m_face_idx;
			m_light_visible = I.m_light_visible;
			m_reflection_info = new intersection_info(*I.m_reflection_info);
			m_refraction_info = new intersection_info(*I.m_refraction_info);
		}
		return *this;
	}

	inline const model *get_model() const {return m_model;}
	inline const int get_face_idx() const {return m_face_idx;}
	inline unsigned int get_light_visible() const {return m_light_visible;}
	inline bool get_light_visible(unsigned int idx) const {return ((get_light_visible() & (1 << idx)) != 0);}

	inline const intersection_info *get_reflection_info() const {return m_reflection_info;}
	inline const intersection_info *get_refraction_info() const {return m_refraction_info;}

private:
	model *m_model;		 ///< the model the ray intersects (NULL if none)
	int m_face_idx;			///< the face the ray intersects 
	unsigned int m_light_visible; // a bitfield (bit 0 = light 0, bit 1 = light 1, etc)

	intersection_info *m_reflection_info; ///< NULL if no reflection or no model is hit
	intersection_info *m_refraction_info; ///< NULL if no refraction or no model is hit
};


#endif /* _INTERSECTION_INFO_H_ */

